#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <unistd.h>

#include "ctar.h"

int create(char *fname)
{
    struct stat statBuf;
    union TarInfo tarInfo;
    int nameLen;
    char nameBuf[NAME_FIELD_SIZE];
    char dnameBuf[NAME_FIELD_SIZE];
    extern char **exclude, **hrdlink;
    extern ino_t *inode;
    extern int numXfiles, numHlinks, numArgc;
    char *p;
    register int i, ec, len, linkidx=0;
    int isdir=FALSE, islnk=FALSE, ischr=FALSE, 
        isblk=FALSE, isfifo=FALSE, isreg=FALSE, 
	ishlnk=FALSE;
    
/* skip "." and ".." */
      if ((strcmp(fname,"..") == 0) || (strcmp(fname,".") == 0)) {
#ifdef DEBUG
fprintf(stderr,"create(): skipping dotdot\n");
#endif
        return(0);
      }
/* return if it's a switch */
      if ((strcmp(fname,"-x") == 0) || (strcmp(fname,"-X") == 0) || (strcmp(fname,"--exclude") == 0))
      {
#ifdef DEBUG
fprintf(stderr,"create(): skipping -x\n");
#endif
        return(0);
      }
/* return if name is in the exclude list */
      for(i=0;i<numXfiles;i++) {
#ifdef DEBUG
fprintf(stderr,"create(): comparing file %s with excluded file %s\n",fname,exclude[i]);
#endif
        if(strcmp(fname,exclude[i]) == 0) {
#ifdef DEBUG
fprintf(stderr,"create(): skipping excluded file %s\n",exclude[i]);
#endif
	  return(0);
        }
      }
/* return if name > 99 chars */
      if ((nameLen=strlen(fname)) >= NAME_FIELD_SIZE) return(E_NAMETOOLONG);
/* stat the file */
      if(lstat(fname,&statBuf) != 0) {
#ifdef DEBUG
fprintf(stderr,"create(): lstat'ing %s\n",fname);
#endif
        perror("Can't stat");
	return(E_CANNOTSTAT);
      }
/* zero the buffers and save a copy of file name */
      memset(nameBuf,0,(size_t) NAME_FIELD_SIZE);
      memset(dnameBuf,0,(size_t) NAME_FIELD_SIZE);
      memset(&tarInfo,0,(size_t) BLOCKSIZE);
      strncpy(dnameBuf,fname,(size_t) nameLen);
/* strip leading '/' */  
      while (*fname == '/') fname++; 
/* if filename was just '/' fname is an empty string now */
      if(!strcmp(fname,"\0")) {
        if(S_ISDIR(statBuf.st_mode)) {	/* make sure it's '/' */
          if((ec=dump_dir(dnameBuf)) != 0) return(ec);
	}
	else return(E_STRANGE_ERR);
      }
/* see if there's more than 1 link to this file */      
#ifdef DEBUG
fprintf(stderr,"create(79): file %s has %d hardlinks\n",fname,(int)statBuf.st_nlink);
fprintf(stderr,"create(80): num hlinks = %d, hlinks = %p\n", numHlinks, hrdlink);
#endif
      if(((int)statBuf.st_nlink > 1) && (!S_ISDIR(statBuf.st_mode))){
/* if list of hardlinks is not empty */
        if(numHlinks > 0) {
          for(i=0;i<numHlinks;i++) { 
#ifdef DEBUG
fprintf(stderr,"create(87): comparing possible hardlink %s with file %s\n",fname,hrdlink[i]);
fprintf(stderr,"create(88): hardlink[%d] at %p is %s\n",i,hrdlink[i], hrdlink[i]);
#endif
            if(inode[i] == statBuf.st_ino)
	    {
#ifdef DEBUG
fprintf(stderr,"create(): adding file %s as hardlink to %s (hrdlink[%d])\n",fname,hrdlink[i],i);
#endif
              ishlnk=TRUE;
	      linkidx = i;
            }
	  }
	  if(!ishlnk) { /* i.e. not found in the list */
#ifdef DEBUG
fprintf(stderr,"create(): adding file %s to list of hardlinks(%d)\n",fname,numHlinks);
#endif
            len = strlen(fname);
            if((hrdlink[numHlinks]=(char *)malloc((size_t)((len + 1)*sizeof(char *)))) == NULL)
            {
              fprintf(stderr,"malloc() error in create(), hrdlink[%d]",numHlinks);
              exit(1);
            }
	    if((inode=(ino_t *)realloc(inode,(size_t)(numHlinks + 1)*sizeof(ino_t))) == NULL)
	    {
              fprintf(stderr,"realloc() error in create(), inode[%d]",numHlinks);
              exit(1);
            }
            strncpy(hrdlink[numHlinks],fname,len);
	    hrdlink[numHlinks][len] = '\0';
            inode[numHlinks] = statBuf.st_ino; 
	    numHlinks++;
          }
	}
	else { /* numHlinks == 0, list is empty */
#ifdef DEBUG
fprintf(stderr,"create(): creating list of hardlinks, adding file %s\n",fname);
#endif
/* 
 * number of hardlinks can't be > argc - 1 
 * (argc - 1 if all args are hardlinks to ctar executable :)
 */
          if((hrdlink=(char **)malloc((size_t)((numArgc - 1 - numXfiles)*sizeof(char *)))) == NULL)
          {
            fprintf(stderr,"malloc() error in create(), hrdlink");
            exit(1);
          }
	  if((inode=(ino_t *)calloc(1, (size_t)sizeof(ino_t))) == NULL) 
	  {
            fprintf(stderr,"calloc() error in create(), inode");
            exit(1);
          }
          len = strlen(fname);
          if((hrdlink[0]=(char *)malloc((size_t)((len + 1)*sizeof(char *)))) == NULL)
          {
            fprintf(stderr,"malloc() error in create(), hrdlink[0]");
            exit(1);
          }
          strncpy(hrdlink[0],fname,len);
	  hrdlink[0][len] = '\0';
	  inode[0] = statBuf.st_ino; 
          numHlinks=1;
        }
      }  /* endif link count > 1 */
      
/* ok, fun part.  We have to fill in the fields of tar header in order,
 * otherwise the header gets corrupted.  I suspect we've gcc's optimizations
 * to thank for that...  or maybe I'm just mising something obvious.
 */
      if(S_ISDIR(statBuf.st_mode)) isdir=TRUE;		/* a directory */
      else if(S_ISLNK(statBuf.st_mode)) islnk=TRUE;	/* link */
      else if(S_ISCHR(statBuf.st_mode)) ischr=TRUE;	/* character device */
      else if(S_ISBLK(statBuf.st_mode)) isblk=TRUE;	/* block device */
      else if(S_ISFIFO(statBuf.st_mode)) isfifo=TRUE;	/* named pipe */
      else if(S_ISREG(statBuf.st_mode)) isreg=TRUE;	/* regular file */
      else if(S_ISSOCK(statBuf.st_mode)) isfifo=TRUE;	/* socket */
      else /* WTF ? */ return(E_WRONG_FTYPE);

      if(isdir) {
/* make sure we have single '/' after the name */
        strncpy(nameBuf, fname, (size_t) nameLen);
	i=nameLen;
	while(nameBuf[i] == '\0') i--;
	while(nameBuf[i] == '/') i--;
	nameBuf[++i] = '/';
	i=nameLen;
	while(dnameBuf[i] == '\0') i--;
	while(dnameBuf[i] == '/') i--;
	dnameBuf[++i] = '/';
#ifdef DEBUG
fprintf(stderr,"create(): fname=%s, nameBuf = %s, nameLen = %d, dnameBuf = %s\n",fname,nameBuf,nameLen,dnameBuf);
#endif
/*
 * at this point we have path without leading '/' in const char *fname
 * -- we can't edit fname without turning -fwritable-strings on,
 * same with single trailing '/' in string nameBuf -- that's the path we 
 * _store_in_the_header, and original path with single trailing '/' in 
 * dnameBuf -- need that to handle absolute paths -- that's the path we 
 * _pass_to_either dump_file() or dump_dir()
 */
        strncpy(tarInfo.header.name,nameBuf,(size_t)NAME_FIELD_SIZE);
      }
      else strncpy(tarInfo.header.name,fname,(size_t)NAME_FIELD_SIZE);
      to_oct ((long) statBuf.st_mode, 8, tarInfo.header.mode);
      to_oct ((long) statBuf.st_uid, 8, tarInfo.header.uid);
      to_oct ((long) statBuf.st_gid, 8, tarInfo.header.gid);
      if((isdir) || (islnk) || (ishlnk)) statBuf.st_size = 0;
      to_oct ((long) statBuf.st_size, 1 + 12, tarInfo.header.size);
      to_oct ((long) statBuf.st_mtime, 1 + 12, tarInfo.header.mtime);
      memcpy (tarInfo.header.chksum, CHKBLANKS, sizeof (tarInfo.header.chksum));
      if(!ishlnk) {
        if(isreg) tarInfo.header.typeflag=REGTYPE; 
        else if(isdir) tarInfo.header.typeflag=DIRTYPE;
        else if(ischr) tarInfo.header.typeflag=CHRTYPE;
        else if(isblk) tarInfo.header.typeflag=BLKTYPE;
        else if(isfifo) tarInfo.header.typeflag=FIFOTYPE;
        else if(islnk) tarInfo.header.typeflag=SYMTYPE;
      }
      else tarInfo.header.typeflag=LNKTYPE;
      if(islnk) {
#ifdef DEBUG
fprintf(stderr,"create(): reading link %s\n",dnameBuf);
#endif
        if((ec=readlink(dnameBuf, nameBuf, NAME_FIELD_SIZE)) < 0) {
          perror("readlink() failed");
          return(E_READLINK);
        }
        if(ec >= NAME_FIELD_SIZE) return(E_NAMETOOLONG);
	nameBuf[ec++]='\0';
        strncpy(tarInfo.header.linkname, nameBuf, (size_t) NAME_FIELD_SIZE);
      }
      if(ishlnk) strcpy(tarInfo.header.linkname, hrdlink[linkidx]);
      strncpy (tarInfo.header.magic, TMAGIC, TMAGLEN);
      strncpy (tarInfo.header.version, TVERSION, TVERSLEN);
      uid_to_uname (statBuf.st_uid, tarInfo.header.uname);
      gid_to_gname (statBuf.st_gid, tarInfo.header.gname);
      if((ischr) || (isblk)) {
        to_oct ((long)major(statBuf.st_rdev), 8, tarInfo.header.devmajor);
        to_oct ((long)minor(statBuf.st_rdev), 8, tarInfo.header.devminor);
      }
        
/* compute checksum */  
      memcpy (tarInfo.header.chksum, CHKBLANKS, sizeof (tarInfo.header.chksum));
      p = tarInfo.buf;
      for (ec=0, i = sizeof (tarInfo); --i >= 0; ) ec += 0xFF & *p++;
      sprintf(tarInfo.header.chksum, "%6o", ec);  
      
/* dump it */      
      if(fwrite(tarInfo.buf, (size_t) 1, (size_t) BLOCKSIZE, stdout) < BLOCKSIZE) 
        return(E_SHORTWRITE);
      if(!ishlnk) {
        if(isreg)
          if((ec=dump_file(dnameBuf,&statBuf)) != 0) return(ec);
        if(isdir) {
          if((ec=dump_dir(dnameBuf)) != 0) return(ec);
        }
      }
      return(0);
}


